/*
 * Copyright (C) 2004-2005 Andre Noll <maan@systemlinux.org>
 *
 *     This program is free software; you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation; either version 2 of the License, or
 *     (at your option) any later version.
 *
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License
 *     along with this program; if not, write to the Free Software
 *     Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111, USA.
 */

#include <ogg/ogg.h>
#include <vorbis/codec.h>
#include <vorbis/vorbisfile.h>

#include "server.h"

/* must be big enough to hold header */
#define CHUNK_SIZE 32768
static double chunk_time = 0.2;

static OggVorbis_File *oggvorbis_file;
static FILE *infile;
static int header_len, vi_channels;
static ogg_int64_t *chunk_table, max_chunk_len;
struct audio_format *af;
static long loaded_bytes, vi_sampling_rate, vi_bitrate, vi_bitrate_nominal,
	num_chunks, initial_chunks = 0;

static int compute_header_len(void)
{
	int ret, len, in = fileno(infile);
	unsigned int serial;
	char *buf;
	ogg_page page;
        ogg_packet packet;
	vorbis_comment vc;
	vorbis_info vi;
	ogg_stream_state *stream_in = para_malloc(sizeof(ogg_stream_state));
	ogg_stream_state *stream_out = para_malloc(sizeof(ogg_stream_state));
	ogg_sync_state *sync_in = para_malloc(sizeof(ogg_sync_state));

	ogg_sync_init(sync_in);
	vorbis_info_init(&vi);
	vorbis_comment_init(&vc);
	buf = ogg_sync_buffer(sync_in, CHUNK_SIZE);
	ret = -1;
	len = read(in, buf, CHUNK_SIZE);
	if (len < 0) {
		PARA_ERROR_LOG("read error %d (%s)\n", len, strerror(errno));
		goto err1;
	}
	if (!len) {
		PARA_NOTICE_LOG("%s\n", "early eof");
		goto err1;
	}
	ogg_sync_wrote(sync_in, len);
	if (ogg_sync_pageout(sync_in, &page) <= 0) {
		PARA_ERROR_LOG("%s", "no ogg file!\n");
		goto err1;
	}
	serial = ogg_page_serialno(&page);
	ogg_stream_init(stream_in, serial);
	ogg_stream_init(stream_out, serial);
	ret = ogg_stream_pagein(stream_in, &page);
	if (ret < 0) {
		PARA_ERROR_LOG("%s", "error in first page\n");
		goto err2;
	}
	ret = ogg_stream_packetout(stream_in, &packet);
	if (ret != 1) {
		PARA_ERROR_LOG("%s", "error in first packet\n");
		goto err2;
	}
	if (vorbis_synthesis_headerin(&vi, &vc, &packet) < 0) {
		PARA_ERROR_LOG("%s", "error in primary header: not vorbis?\n");
		goto err2;
	} else
		PARA_INFO_LOG("channels: %i, rate: %li\n", vi.channels,
			vi.rate);
	ogg_stream_packetin(stream_out, &packet);

	ret = ogg_sync_pageout(sync_in, &page);
	ogg_stream_pagein(stream_in, &page);
	ogg_stream_packetout(stream_in, &packet);
	ogg_stream_packetin(stream_out, &packet);

	ret = ogg_sync_pageout(sync_in, &page);
	ogg_stream_pagein(stream_in, &page);
	ogg_stream_packetout(stream_in, &packet);
	ogg_stream_packetin(stream_out, &packet);

	header_len = 0;
	while (ogg_stream_flush(stream_out, &page))
		header_len += page.body_len + page.header_len;
	ret = len;
	PARA_INFO_LOG("header_len = %d\n", header_len);
err2:
	ogg_stream_destroy(stream_in);
	ogg_stream_destroy(stream_out);
err1:
	ogg_sync_destroy(sync_in);
	vorbis_info_clear(&vi);
        vorbis_comment_clear(&vc);
	return ret;
}

/*
 * Alloc and fill array table of byte offsets. chunk_table[i] is the
 * offset in the current infile at which the sample containing time i *
 * CHUNK_TIME begins.
 */
static void ogg_compute_chunk_table(double time_total, ogg_int64_t raw_total)
{
	int i, ret, num;
	ogg_int64_t pos = 0, min = 0, old_pos;

	old_pos = 0;
	ret = 0;
	free(chunk_table);
	num = time_total / chunk_time + 3;
	PARA_DEBUG_LOG("chunk time: %g allocating %d chunk pointers\n",
		chunk_time, num);
	chunk_table = para_malloc(num * sizeof(ogg_int64_t));
	chunk_table[0] = 0;
	max_chunk_len = 0;
	initial_chunks = 0;
	rewind(infile);
	for (i = 1; ret == 0; i++) {
		ogg_int64_t diff;
		ret = ov_time_seek(oggvorbis_file, i * chunk_time);
		if (ret)
			break;
		pos = ov_raw_tell(oggvorbis_file);
		diff = pos - old_pos;
		/* find first i such that chunk_table[i] > chunk_size / 2 */
		if (pos > CHUNK_SIZE / 2 && pos < CHUNK_SIZE &&
				old_pos < CHUNK_SIZE / 2)
			initial_chunks = i - 1;
		max_chunk_len = MAX(max_chunk_len, diff);
		min = (i == 1)? diff : MIN(min, diff);
		chunk_table[i] = pos;
		if (i < 11 || !((i - 1) % 1000)|| i > num - 11)
			PARA_DEBUG_LOG("chunk #%d: %g secs, pos: %lli, "
				"size: %lli\n", i - 1,
				i * chunk_time, pos, pos - old_pos);
		old_pos = pos;
	}
	chunk_table[i] = pos;
	num_chunks = i - 1;
	if (initial_chunks > 0) {
		for (i = 1; i < initial_chunks; i++) {
			chunk_table[i] = chunk_table[initial_chunks];
		}
		max_chunk_len = MAX(max_chunk_len, chunk_table[initial_chunks]);
	}
	PARA_INFO_LOG("%li chunks (%fs), max chunk: %lli, min chunk: %lli, "
		"initial_chunks: %li\n", num_chunks, chunk_time,
		max_chunk_len, min, initial_chunks);
	rewind(infile);
}

static void ogg_close_ov_file(void)
{
	if (oggvorbis_file) {
		PARA_DEBUG_LOG("%s", "ov_clear\n");
		ov_clear(oggvorbis_file);
		oggvorbis_file = NULL;
	}
}

static void ogg_close_fifo(int fifo[2])
{
	if (fifo[1] > 0) {
		PARA_INFO_LOG("closing write fd %d\n", fifo[1]);
		close(fifo[1]);
		af->fifo = -1;
	}
	if (fifo[0] > 0) {
		PARA_INFO_LOG("closing read fd %d\n", fifo[0]);
		close(fifo[0]);
	}
}

/*
 * Init oggvorbis file and write some tech data to given pointers.
 */
static int ogg_get_file_info(FILE *file, char *info_str, long unsigned *frames,
	int *seconds)
{
	int ret;
	double time_total;
	vorbis_info *vi;
	ogg_int64_t raw_total;

	infile = file;
	if (!file) {
		PARA_CRIT_LOG("%s", "invalid file pointer\n");
		return -1;
	}
	ret = compute_header_len();
	if (ret < 0)
		return -1;
	rewind(file);
	oggvorbis_file = para_malloc(sizeof(OggVorbis_File));
	ret = ov_open(file, oggvorbis_file, NULL, 0);
	if (ret < 0) {
		PARA_ERROR_LOG("%s", "ogg format not recognized\n");
		free(oggvorbis_file);
		return ret;
	}
	ret = -1;
	vi = ov_info(oggvorbis_file, 0);
	if (!vi)
		goto err;
	time_total = ov_time_total(oggvorbis_file, -1);
	raw_total = ov_raw_total(oggvorbis_file, -1);
	*seconds = time_total;
	vi_sampling_rate = vi->rate;
	vi_bitrate = ov_bitrate(oggvorbis_file, 0);
	vi_bitrate_nominal = vi->bitrate_nominal;
	vi_channels = vi->channels;
	ogg_compute_chunk_table(time_total, raw_total);
	*frames = num_chunks;
	sprintf(info_str, "audio_file_info1:%ldkHz, %d channels, %ldkbps, "
		"%lu chunks (%g ms)\n"
		"audio_file_info2: \n"
		"audio_file_info3: \n",
		vi_sampling_rate / 1000, vi_channels, vi_bitrate / 1000,
		num_chunks, chunk_time * 1000);
	rewind(file);
	return 1;
err:
	ogg_close_ov_file();
	return ret;
}

/*
 * Simple stream reposition routine
 */
static int ogg_repos(long unsigned request)
{
	int ret;
	long seek;

	seek = chunk_table[request];
	ret = fseek(infile, seek, SEEK_SET);
	PARA_DEBUG_LOG("seek to %li returned %d. offset:%li\n", seek,
		ret, ftell(infile));
	if (ret)
		goto out;
out:
	return ret;
}

static ogg_int64_t get_chunk_size(long unsigned chunk_num)
{
	ogg_int64_t ret;
	if (chunk_num >= num_chunks)
		return -1;
	ret = chunk_table[chunk_num + 1] - chunk_table[chunk_num];
	if (!ret)
		return ret;
#if 0
	PARA_DEBUG_LOG("chunk %d: %lli-%lli (%lli bytes)\n",
		chunk_num,
		chunk_table[chunk_num],
		chunk_table[chunk_num + 1] - 1,
		ret);
#endif
	return ret;
}

/*
 * when resuming after pause we have to resent the header.
 */
static int send_extra_header(void)
{
	long cur_pos;
	int bytes, ret = -1;
	char *buf = para_malloc(header_len * sizeof(char));

	if (!header_len || !oggvorbis_file)
		goto out;
	cur_pos = ftell(infile); /* remember current position */
	if (cur_pos == -1)
		goto out;
	rewind(infile);
        bytes = para_fread(buf, header_len, 1, infile);
	if (bytes != header_len) {
		PARA_ERROR_LOG("fread returned %d, expected: %d\n",
			bytes, header_len);
		ret = loaded_bytes;
		goto out;
	}
	PARA_DEBUG_LOG("sending extra header (%d bytes)\n", header_len);
	ret = write(af->fifo, buf, header_len);
	if (ret != header_len)
		ret = -1;
	ret = fseek(infile, cur_pos, SEEK_SET);
out:
	return ret;
}

static int ogg_send_chunk(char *info_str, long unsigned chunk_num,
	long unsigned chunks_sent, int afs_status)
{
	static int buflen;
	int ret, err;
	static char *buf;
	ogg_int64_t chunk_size;

	if (loaded_bytes > 0) {
		PARA_INFO_LOG("completing short write (%li loaded bytes not "
			"yet written)\n", loaded_bytes);
		goto write;
	}
	if (!buf || buflen < max_chunk_len) {
		PARA_INFO_LOG("increasing ogg buffer size (%d -> %lli)\n",
			buflen, max_chunk_len);
		buf = para_realloc(buf, max_chunk_len);
		buflen = max_chunk_len;
	}
	if (!chunks_sent && chunk_num) {
		/* we just resumed from pause */
		ret  = send_extra_header();
		if (ret < 0)
			PARA_CRIT_LOG("failed to send extra header (%d)\n",
				ret);
	}
	chunk_size = get_chunk_size(chunk_num);
	if (!chunk_size) /* nothing to send for this run */
		return 1;
	if (chunk_size < 0)
		return 0;
	if (chunk_num < 10 || !(chunk_num % 1000)
			|| chunk_num + 10 > num_chunks)
		PARA_DEBUG_LOG("reading chunk #%li (%lli bytes). Sent: %li\n",
			chunk_num, chunk_size, chunks_sent);
        loaded_bytes = para_fread(buf, chunk_size, 1, infile);
	if (loaded_bytes <= 0) {
		PARA_DEBUG_LOG("read #%li returned %li, expected: %lli\n",
			chunk_num, loaded_bytes, chunk_size);
		ret = loaded_bytes;
		loaded_bytes = 0;
		return ret;
	}
//	PARA_DEBUG_LOG("%x:%x:%x\n", buf[0], buf[1], buf[2]);
	if (loaded_bytes && loaded_bytes < chunk_size) {
		PARA_CRIT_LOG("oops, short read (%li/%lli)\n",
			loaded_bytes, chunk_size);
		loaded_bytes = 0;
		return -1;
	}
write:
	ret = write(af->fifo, buf, loaded_bytes);
	err = errno;
	if (ret <= 0) {
		PARA_INFO_LOG("write failed (%s), %d\n", strerror(err), err);
		loaded_bytes = 0;
		return ret;
	}
	if (ret < loaded_bytes) {
		PARA_WARNING_LOG("%s", "short write\n");
		memcpy(buf, buf + ret, loaded_bytes - ret);
		loaded_bytes -= ret;
		return 1;
	}
	loaded_bytes -= ret;
	return 1;
}

static char *glob_cmd(void)
{
	char *ret, *replacement = make_message("%d", header_len);

	if (!af->open_cmd) {
		char *hn = para_hostname(), *home = para_homedir();
		af->open_cmd = make_message("para_send --header_length HEADER_LEN() "
			"%s/.paraslash/server_ogg_fifo.%s", home, hn);
		free(hn);
		free(home);
	}
	ret = s_a_r(af->open_cmd, "HEADER_LEN", replacement);
	free(replacement);
	return ret;
}

static int ogg_mainloop(int current_af_is_ogg, unsigned int afs_status,
	struct timeval **tvptr, long unsigned chunk_num,
	long unsigned chunks_sent, long unsigned repos_request, long *offset,
	struct timeval *stream_start, struct timeval *sw_exec, long unsigned *eof_time)
{
	unsigned next = afs_status & AFS_NEXT;
	unsigned playing = afs_status & AFS_PLAYING;
	unsigned paused = !(afs_status & ~AFS_STREAM_WRITER);
	unsigned repos = afs_status & AFS_REPOS;
	long timeout;
	static struct timeval tv;
	struct timeval tmp;
	static int fifo[2] = {-1, -1}, sw[3] = {0, 0, 0};

	*tvptr = NULL;
	if (next || paused || repos) {
		if (af->sw_pid > 0 && af->fifo > 0)
			ogg_close_fifo(fifo);
		if (next)
			ogg_close_ov_file();
		loaded_bytes = 0;
		if (repos && chunk_num != repos_request && current_af_is_ogg)
			ogg_repos(repos_request);
		return -1;
	}
	if (!current_af_is_ogg)
		return -1;
	if (playing && af->sw_pid <= 0 && oggvorbis_file) {
		char *glob = glob_cmd();
		int ret;
		*eof_time = 1000 * (initial_chunks + 3) * chunk_time;
		if (*eof_time > 4000)
			*eof_time = 4000;
		ret = para_exec_cmdline_pid(&af->sw_pid, glob, sw);
		gettimeofday(sw_exec, NULL);
		PARA_DEBUG_LOG("execed %s\n", glob);
		free(glob);
		if (ret < 0)
			return -1;
		if (af->fifo < 0) {
			if (afs_open_fifo(af->fifoname, fifo) < 0)
				return -1;
			af->fifo = fifo[1];
		}
		*offset = chunk_num * 1000 * chunk_time;
		return af->fifo;
	}

	if (!playing || !oggvorbis_file || af->fifo < 0 || !af->sw_pid)
		return -1;
	timeout = chunks_sent * chunk_time * 1000 - *eof_time / 2;
	if (timeout > 1000) {
		ms2tv(timeout - 1000, &tmp);
		tv_add(&tmp, stream_start, &tv);
		*tvptr = &tv;
	}
	return af->fifo;
}

int ogg_init(void *p)
{
	af = p;
	af->get_file_info = ogg_get_file_info,
	af->send_chunk = ogg_send_chunk;
	af->mainloop = ogg_mainloop;
	af->fifo = -1;
	return 1;
}
